//Copyright (c) 2003, DPeshev
//All rights reserved.
//
//Redistribution and use in source and binary forms, with or without modification, 
//are permitted provided that the following conditions are met:
//
//- Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer. 
//
//- Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution. 
//
//- Neither the name of the <ORGANIZATION> nor the names of its contributors may 
//  be used to endorse or promote products derived from this software without 
//  specific prior written permission. 
//
//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
//LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
//DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
//SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
//IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//POSSIBILITY OF SUCH DAMAGE.

#undef NDEBUG
#include <assert.h>
#include "stdafx.h"
#include "AudioBase.h"


using namespace std;

class CAudioPluginMgr : public IAudioPluginMgr, public IProfile {
public:
	CAudioPluginMgr() : m_lRef(0) {};
	virtual ~CAudioPluginMgr() {}
public:
	// IUnknown
	STDMETHODIMP QueryInterface(REFIID riid, void __RPC_FAR* __RPC_FAR* ppObj);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();
	// IAudioPluginMgr
	virtual bool SetPluginDirectory(const char *szPluginDir);
	virtual bool LoadPlugins();
	virtual bool CanUnloadPlugins();
	virtual bool UnloadPlugins();
	virtual bool SetConvertor(IConvertorFactory *pCF, IStatus **ppStatus = NULL);
	virtual bool GetConvertor(IConvertorFactory **ppConvFactory);
	virtual bool GetUnsupportedComponents(IComponentEnum **ppEnum);
	virtual bool SendEvent(EEvent event, IUnknown *pSource);
	virtual bool SubscribeToEvents(IEventReceiver *pReceiver);
	virtual bool UnsubscribeToEvents(IEventReceiver *pReceiver);
	virtual bool OpenURLAudioFile(	const char			*szURL,
						const SWavFormat	*pFormat,
						IUnknown			**ppSrc,
						int					iTries,
						EAuxFlags			flagsInclude,
						EAuxFlags			flagsExclude,
						IStatus				**ppStatus);
	virtual bool CreateURLAudioTarget(	IURLAudioTargetFactory	*pFactory,
							IUnknown				**pTgt,
							const SWavFormat		&formatSrc,
							IStatus					**ppStatus	= NULL);
	virtual bool AttachAuxObject(REFCLSID clsID, IUnknown *pUnknown);
	virtual bool DetachAuxObject(REFCLSID clsID);
	virtual bool GetAuxObject(REFCLSID clsID, IUnknown **ppUnknown);
	virtual bool TranslateError(IStatus *pInStatus, IStatus **ppOutStatus);
	// IProfile
	virtual bool SetVendor(const char *szVendor, IStatus **ppStatus = NULL);
	virtual const char *GetVendor();
	virtual bool SetProduct(const char *szProduct, IStatus **ppStatus = NULL);
	virtual const char * GetProduct();
	// Data
protected:
	long    m_lRef;
};
//IAudioPluginMgr
bool CAudioPluginMgr::SetPluginDirectory(const char *szPluginDir) {return true;};
bool CAudioPluginMgr::LoadPlugins() {return true;};
bool CAudioPluginMgr::CanUnloadPlugins() {return true;};
bool CAudioPluginMgr::UnloadPlugins() {return true;};
bool CAudioPluginMgr::SetConvertor(IConvertorFactory *pCF, IStatus **ppStatus) {return true;};
bool CAudioPluginMgr::GetConvertor(IConvertorFactory **ppConvFactory) {return true;};
bool CAudioPluginMgr::GetUnsupportedComponents(IComponentEnum **ppEnum) {return true;};
bool CAudioPluginMgr::SendEvent(EEvent event, IUnknown *pSource) {return true;};
bool CAudioPluginMgr::SubscribeToEvents(IEventReceiver *pReceiver) {return true;};
bool CAudioPluginMgr::UnsubscribeToEvents(IEventReceiver *pReceiver) {return true;};
bool CAudioPluginMgr::OpenURLAudioFile(	const char			*szURL,
						const SWavFormat	*pFormat,
						IUnknown			**ppSrc,
						int					iTries,
						EAuxFlags			flagsInclude,
						EAuxFlags			flagsExclude,
						IStatus				**ppStatus) {return true;};
bool CAudioPluginMgr::CreateURLAudioTarget(	IURLAudioTargetFactory	*pFactory,
							IUnknown				**pTgt,
							const SWavFormat		&formatSrc,
							IStatus					**ppStatus) {return true;};
bool CAudioPluginMgr::AttachAuxObject(REFCLSID clsID, IUnknown *pUnknown) {return true;};
bool CAudioPluginMgr::DetachAuxObject(REFCLSID clsID) {return true;};
bool CAudioPluginMgr::GetAuxObject(REFCLSID clsID, IUnknown **ppUnknown) {return true;};
bool CAudioPluginMgr::TranslateError(IStatus *pInStatus, IStatus **ppOutStatus) {return true;};
// IProfile
bool CAudioPluginMgr::SetVendor(const char *szVendor, IStatus **ppStatus) {return true;};
const char *CAudioPluginMgr::GetVendor() { return "Smeshka";};
bool CAudioPluginMgr::SetProduct(const char *szProduct, IStatus **ppStatus) {return true;};
const char * CAudioPluginMgr::GetProduct() { return "Nencode"; };
// QueryInterface returns either IUnknown, IAudioPluginMgr or IProfile
STDMETHODIMP CAudioPluginMgr::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppObj) {
  // ppObj must not be NULL
  if(NULL == ppObj) {
    assert(FALSE);
    return E_INVALIDARG;
  }
  *ppObj = NULL;
  if(IsEqualIID(riid, IID_IUnknown)) *ppObj = static_cast<IUnknown*>(static_cast<IAudioPluginMgr*>(this));
  else if(IsEqualIID(riid, IID_IAudioPluginMgr)) *ppObj = static_cast<IAudioPluginMgr*>(this);
  else if(IsEqualIID(riid, IID_IProfile)) *ppObj = static_cast<IProfile*>(this);
  if(*ppObj) AddRef();
  if (NULL != *ppObj) return S_OK;
  else return E_NOINTERFACE;
}

// Increase the reference counter and return the number of references
STDMETHODIMP_(ULONG) CAudioPluginMgr::AddRef() {
  InterlockedIncrement(&m_lRef);
  return m_lRef;
}

// Decrease the reference counter and return the number of references
STDMETHODIMP_(ULONG) CAudioPluginMgr::Release() {
  InterlockedDecrement(&m_lRef);
  return m_lRef;
}

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[]) {
	int nRetCode = 0;
	CAudioPluginMgr pluginMgr;
	IStatus *pIStatus;
	typedef bool (* LPNERO_PLUGIN_GetPrimaryAudioObject)(IAudioComponent**);
	typedef bool (* LPNERO_PLUGIN_ReadyToFinish)();

	HINSTANCE hWavDLL,hAacDLL;            
	LPNERO_PLUGIN_GetPrimaryAudioObject wavGetPrimaryAudioObject,aacGetPrimaryAudioObject;
	LPNERO_PLUGIN_ReadyToFinish wavReadyToFinish,aacReadyToFinish;
	IAudioComponent *pWavIA,*pAacIA;

	hWavDLL = LoadLibrary("wav"); assert(hWavDLL);
	hAacDLL = LoadLibrary("aac"); assert(hAacDLL);
		
	wavGetPrimaryAudioObject = 
		(LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress(hWavDLL,"NERO_PLUGIN_GetPrimaryAudioObject");
	assert(wavGetPrimaryAudioObject);
	wavReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress(hWavDLL,"NERO_PLUGIN_ReadyToFinish");
	assert(wavReadyToFinish);
	aacGetPrimaryAudioObject = 
		(LPNERO_PLUGIN_GetPrimaryAudioObject)GetProcAddress(hAacDLL,"NERO_PLUGIN_GetPrimaryAudioObject");
	assert(aacGetPrimaryAudioObject);
	aacReadyToFinish = (LPNERO_PLUGIN_ReadyToFinish)GetProcAddress(hAacDLL,"NERO_PLUGIN_ReadyToFinish");
	assert(aacReadyToFinish);
	nRetCode = wavGetPrimaryAudioObject(&pWavIA); 
	assert(nRetCode && pWavIA->GetType() == ACT_ComponentEnumerator);
	nRetCode = aacGetPrimaryAudioObject(&pAacIA); 
	assert(nRetCode && pAacIA->GetType() == ACT_ComponentEnumerator);
	IConfigurable *pIC;
	IComponentEnum *pWavICU,*pAacICU; 
	nRetCode = pWavIA->QueryInterface(IID_IComponentEnum,(void **)&pWavICU);
	assert(!nRetCode && pWavICU && pWavICU->GetCount()>0);
	nRetCode = pAacIA->QueryInterface(IID_IComponentEnum,(void **)&pAacICU);
	assert(!nRetCode && pAacICU && pAacICU->GetCount()>1);
	IAudioComponent *pWavIAC,*pAacIAC;
	nRetCode = pWavICU->GetComponent(0,&pWavIAC);
	assert(nRetCode && pWavIAC && pWavIAC->GetType() == ACT_AudioURLSourceFactory);
	nRetCode = pAacICU->GetComponent(1,&pAacIAC);
	assert(nRetCode && pAacIAC && pAacIAC->GetType() == ACT_AudioURLTargetFactory);
	IURLAudioSourceFactory *pWavSrc;
	nRetCode = pWavIAC->QueryInterface(IID_IURLAudioSourceFactory, (void**)&pWavSrc);
	assert(!nRetCode && pWavSrc);
	IURLAudioTargetFactory *pAacDest;
	nRetCode = pAacIAC->QueryInterface(IID_IURLAudioTargetFactory, (void**)&pAacDest);
	assert(!nRetCode && pAacDest);
	nRetCode = pAacIAC->QueryInterface(IID_IConfigurable, (void**)&pIC);
	assert(!nRetCode && pIC);
	nRetCode = pAacIAC->Init(static_cast<IAudioPluginMgr*>(&pluginMgr), &pIStatus);
	assert(nRetCode && !pIStatus);
	if (argc < 3) {
		printf("usage: nencode [<source.wav> <dest.mp4>] \r\n");
		printf("\r\n");
		printf("               run without parameters to configure MPEG-4 plugin\r\n\r\n");
		printf("<- %s\n\r",pWavIAC->GetName());
		printf("-> %s\n\r",pAacIAC->GetName());
		nRetCode = pIC->Configure(); assert(nRetCode);
		nRetCode = pAacIAC->Done(); assert(nRetCode);
		return 0;
	}
	IUnknown *pIUWav,*pIUAac;
	TCHAR fullSourcePath[_MAX_PATH];
    assert(_tfullpath( fullSourcePath, argv[1], _MAX_PATH ) != NULL);
	nRetCode = pWavSrc->Open(fullSourcePath,&pIUWav);

	assert(nRetCode && pIUWav);
	IAudioRawReader *pWavIAURR;
	nRetCode = pIUWav->QueryInterface(IID_IAudioRawReader, (void**)&pWavIAURR);
	assert(!nRetCode && pWavIAURR);


	SWavFormat format = pWavIAURR->GetRawFormat();

	nRetCode = pAacDest->CreateURLAudioTarget(&pIUAac,format,&pIStatus);
	assert(nRetCode && !pIStatus);
	IAudioRawWriter *pAacIAURW;
	nRetCode = pIUAac->QueryInterface(IID_IAudioRawWriter, (void**)&pAacIAURW);
	assert(!nRetCode && pAacIAURW);
	IInfoWriter *pAacIW;
	nRetCode = pIUAac->QueryInterface(IID_IInfoWriter, (void**)&pAacIW);
	assert(!nRetCode && pAacIAURW);
										
	IURLHolder *pIURLH;
	nRetCode = pIUAac->QueryInterface(IID_IURLHolder, (void**)&pIURLH);
	assert(!nRetCode && pIURLH);
	TCHAR fullDestPath[_MAX_PATH];
    assert(_tfullpath( fullDestPath, argv[2], _MAX_PATH ) != NULL);
	nRetCode = pIURLH->SetURL(fullDestPath,&pIStatus);
	assert(nRetCode && !pIStatus);
	IProcess *pAacIProcess,*pWavIProcess;
	nRetCode = pIUAac->QueryInterface(IID_IProcess, (void**)&pAacIProcess);
	assert(!nRetCode && pAacIProcess);
	nRetCode = pIUWav->QueryInterface(IID_IProcess, (void**)&pWavIProcess);
	assert(!nRetCode && pWavIProcess);
	nRetCode = pAacIProcess->Start(&pIStatus); assert(nRetCode && !pIStatus);
	nRetCode = pWavIProcess->Start(&pIStatus); assert(nRetCode && !pIStatus);
	BYTE pBuf[65000];
	int blockSize, blockSizeRead,starCount = 0;
	enum EAudioRawState state;
	ULONGLONG rawSize = pWavIAURR->GetRawLen();
	ULONGLONG size = rawSize;
	while (size > 0) {
		blockSize = size>65000?65000:(int)size;
		nRetCode = pWavIAURR->RawRead(pBuf,blockSize,&blockSizeRead,state,&pIStatus);
		assert(nRetCode && !pIStatus);
		nRetCode = pAacIAURW->RawWrite(pBuf,blockSizeRead,state,&pIStatus);
		assert(nRetCode && !pIStatus);
		size -= blockSize;
		while (((rawSize-size)*30)/rawSize > starCount) { printf("*"); starCount++; }
	}

	nRetCode = pAacIProcess->End(&pIStatus); assert(nRetCode && !pIStatus);
	nRetCode = pWavIProcess->End(&pIStatus); assert(nRetCode && !pIStatus);
	aacReadyToFinish();
	wavReadyToFinish();
	return 0;
}
